#!/usr/bin/env perl
# Copyright (C) 2019-2020 Keith A. Bare II
# Written for use in the CMU Computer Club's operations.

#
# The purpose of this script is to deal with daemons that need to run with
# Kerberos tickets and/or AFS tokens.
#
# There are multiple tools that attempt to provide support for running a
# service with these sorts of credentials.  E.g., Heimdal kinit will run a
# program as a child and keep Kerberos-based credential up to date as the
# program runs.  The k5start package also provides similar functionality.
# The problem with both is that they only keep the tickets/tokens up to date
# while the child they spawn is running.  Real daemons terminate their
# initial process and keep working in the background--which results in those
# tools thinking their work is done and destroying any credentials they've
# fetched.
#
# This script acts as a bridge that will allow kinit to maintain tickets/
# tokens for real daemons.  Overall, the following processes are inovlved:
#
#  - kdaemonwrap.pl (parent)
#     |
#     +- kinit
#         |
#         +- kdaemonwrap.pl (child)
#             |
#             +- daemon (frontend)
#                 |
#                 +- daemon (worker)
#
# Once everything is started, the process diagram looks as follows in steady
# state.
#
#  - kinit
#     |
#     +- kdaemonwrap.pl (child)
#
#  - daemon (worker)
#
# The kdaemonwrap.pl parent looks like a real daemon (i.e., runs in the
# background) for the purposes of the init system, while the child stays in
# the foreground as long as the wrapped daemon is running, causing kinit to
# continue to think it has work to do and refresh the tickets/tokens.
#

use strict;
use warnings;

use FindBin ();
use Getopt::Long ();
use POSIX ();

my $keytab;
my $pid_file;
my $pid_timeout = 10;
my $parent_pid;
my $principal;
my $tkt_forwardable;
my $tkt_lifetime;
my $tkt_not_forwardable;
my $tkt_proxiable;
my $tkt_renewable_lifetime;
my $verbose = 0;

#
# Restores default signal handling for SIGCHLD and SIGUSR1.
#
sub restore_signal_defaults
{
	$SIG{CHLD} = 'DEFAULT';
	$SIG{USR1} = 'DEFAULT';
}

#
# Installs signal handlers for SIGCHLD and optionally SIGUSR1, first masking
# the signals to ensure the caller can use sigsuspend() without issue.
#
# First (required) argument is a reference to a scalar that will be assigned
# a true value when SIGCHLD is received.
#
# Second (optional) argument is a reference to a scalar that will be
# assigned a true value when SIGUSR1 is received.  A signal handler will not
# be installed for SIGUSR1 if the second argument is omitted.
#
sub install_signal_handlers
{
	my ($got_sigchld_ref, $got_sigusr1_ref) = @_;

	my $sigset = POSIX::SigSet->new();
	POSIX::sigprocmask(&POSIX::SIG_SETMASK, undef, $sigset);

	$sigset->addset(&POSIX::SIGCHLD);
	if (defined($got_sigusr1_ref)) {
		$sigset->addset(&POSIX::SIGUSR1);
	}
	POSIX::sigprocmask(&POSIX::SIG_SETMASK, $sigset, undef);

	$SIG{CHLD} = sub { $$got_sigchld_ref = 1 };
	if (defined($got_sigusr1_ref)) {
		$SIG{USR1} = sub { $$got_sigusr1_ref = 1 };
	}

	$sigset->delset(&POSIX::SIGCHLD);
	if (defined($got_sigusr1_ref)) {
		$sigset->delset(&POSIX::SIGUSR1);
	}

	return $sigset;
}

#
# Routine implementing the parent process logic.  The parent is responsible
# for spawning a kinit process that then re-runs this script with applicable
# arguments in child mode.  After the child is spawned, the parent waits for
# kinit to terminate, or for SIGUSR1 from the child.  If kinit terminates,
# the parent will exit with a status describing kinit's termination.  If
# SIGUSR1 is received, the parent will exit with status 0, allowing kinit
# and the child to continue running as daemon processes.
#
sub parent_handling
{
	$parent_pid = $$;
	if ($verbose >= 2) {
		print STDERR ("Executing parent_handling(), \$parent_pid=",
		              "$parent_pid\n");
	}

	my $kinit_ended = 0;
	my $daemon_started = 0;
	my $sigset = install_signal_handlers(\$kinit_ended, \$daemon_started);

	my $kinit_pid = fork();
	if ($kinit_pid < 0) {
		die "Failed to fork(): $!";
	} elsif ($kinit_pid == 0) {
		restore_signal_defaults();
		execute_kinit();
	}

	if ($verbose >= 2) {
		print STDERR ("Forked, \$kinit_pid=$kinit_pid\n");
	}

	while (!$kinit_ended && !$daemon_started) {
		POSIX::sigsuspend($sigset);
	}

	if ($daemon_started) {
		if ($verbose >= 1) {
			print STDERR ("Daemon started, exiting with status 0\n");
		}
		exit(0);
	}

	my $tries;
	my $kinit_status;
	for ($tries = 3; $tries > 0; $tries--) {
		if (waitpid($kinit_pid, &POSIX::WNOHANG) > 0) {
			$kinit_status = $?;
			last;
		}
		sleep(1);
	}

	if ($tries == 0) {
		die "Could not obtain the child process status";
	}

	if (POSIX::WIFSIGNALED($kinit_status)) {
		my $termsig = POSIX::WTERMSIG($kinit_status);
		print STDERR ("Child process terminated with signal ",
		              $termsig, "\n");
		# This is the same convention kinit uses
		exit(128 + $termsig);
	} else {
		my $exitcode = POSIX::WEXITSTATUS($kinit_status);
		if (($exitcode != 0) || ($verbose >= 1)) {
			print STDERR ("Child process exited with status ",
			              $exitcode, "\n");
		}
		exit($exitcode);
	}
}

#
# Runs kinit via an exec() call.  Builds arguments to kinit based on those
# that were provided to the parent process.  The kinit will then re-execute
# this script, in child mode.
#
sub execute_kinit
{
	my @kinit_args = ("kinit",);

	if (defined($keytab)) {
		push(@kinit_args, "--keytab=$keytab");
	}
	if ($tkt_forwardable) {
		push(@kinit_args, "--forwardable");
	}
	if (defined($tkt_lifetime)) {
		push(@kinit_args, "--lifetime=$tkt_lifetime");
	}
	if ($tkt_not_forwardable) {
		push(@kinit_args, "--no-forwardable");
	}
	if ($tkt_proxiable) {
		push(@kinit_args, "--proxiable");
	}
	if (defined($tkt_renewable_lifetime)) {
		push(@kinit_args, "--renewable-life=$tkt_renewable_lifetime");
	}

	push(@kinit_args, "--");

	push(@kinit_args, $principal);

	push(@kinit_args, "$FindBin::Bin/$FindBin::Script");
	if ($verbose) {
		push(@kinit_args, "-" . ("v" x $verbose));
	}
	push(@kinit_args, "--child");
	push(@kinit_args, "--parent-process=$parent_pid");
	push(@kinit_args, "--pid-file=$pid_file");
	push(@kinit_args, "--pid-timeout=$pid_timeout");

	push(@kinit_args, "--");

	push(@kinit_args, @ARGV);

	if ($verbose >= 1) {
		print STDERR ("+", join(" ", @kinit_args), "\n");
	}
	exec { $kinit_args[0] } (@kinit_args);
	die "Failed to exec(): $!";
}

#
# Routine implementing the child process logic.  The child runs with the
# credentials that kinit acquires.  It will spawn the executable being
# wrapped and wait for its foreground process to terminate.  If the
# termination condition indicates failure, the child will exit with a status
# indicating how the executable terminated.  If the termination condition
# indicates success, the child will wait for the daemon's pid file to
# appear.  When the pid file appears, the child will send SIGUSR1 to the
# parent process, notifying it that the daemon has started.  At that point,
# the child will repeatedly probe the daemon's pid, up until the point that
# it detects the daemon is no longer running.
#
sub child_handling
{
	if ($verbose >= 2) {
		print STDERR ("Executing child_handling, \$\$=$$, ",
		              "\$parent_pid=$parent_pid\n");
	}

	my $daemon_returned = 0;
	my $sigset = install_signal_handlers(\$daemon_returned);

	my $daemon_pid = fork();
	if ($daemon_pid < 0) {
		die "Failed to fork(): $!";
	} elsif ($daemon_pid == 0) {
		restore_signal_defaults();
		if ($verbose >= 1) {
			print STDERR ("+", join(" ", @ARGV), "\n");
		}
		exec { $ARGV[0] } (@ARGV);
		die "Failed to exec(): $!";
	}

	while (!$daemon_returned) {
		POSIX::sigsuspend($sigset);
	}

	kill('SIGUSR1', $parent_pid);

	if (waitpid($daemon_pid, 0) < 0) {
		die "Error waiting on daemon executable: $!";
	}
	my $daemon_status = $?;

	if (POSIX::WIFSIGNALED($daemon_status)) {
		my $termsig = POSIX::WTERMSIG($daemon_status);
		print STDERR ("Daemon executable terminated with signal ",
		              $termsig, "\n");
		# This is the same convention kinit uses
		exit(128 + $termsig);
	} else {
		my $exitcode = POSIX::WEXITSTATUS($daemon_status);
		if (($exitcode != 0) || ($verbose >= 1)) {
			print STDERR ("Daemon executable exited with status ",
			              $exitcode, "\n");
		}
		if ($exitcode != 0) {
			exit($exitcode);
		}
	}

	my $tries;
	for ($tries = $pid_timeout + 1; $tries > 0; $tries--) {
		if (-e $pid_file) {
			last;
		}
		sleep(1);
	}

	if ($tries == 0) {
		die "Pid file did not appear within $pid_timeout seconds";
	}

	open PIDFILE, '<', $pid_file or die "Could not open pid file";
	$daemon_pid = <PIDFILE>; chomp $daemon_pid;
	close PIDFILE;
	$daemon_pid =~ /^[0-9]+$/ or die "Bad (non-numeric) pid from pid file";
	if ($verbose >= 1) {
		print STDERR ("Daemon process has pid $daemon_pid\n");
	}

	my $probed = 0;
	while (kill(0, $daemon_pid) == 1) {
		$probed = 1;
		sleep(1);
	}
	if (!$!{ESRCH}) {
		die "Could not probe the daemon's status: $!";
	}
	if (!$probed) {
		print STDERR ("Warning: could not confirm daemon process was ",
		              "ever running\n");
	}
	if ($verbose >= 1) {
		print STDERR ("Detected daemon process termination, exiting\n");
	}

	exit(0);
}

#
# Entry point for script execution.  Called both when running as the parent
# process and as the child process.  The routine distinguishes between the
# two cases by checking whether a --child argument was provided in @ARGV.
#
sub main
{
	my $child = 0;

	Getopt::Long::Configure("bundling");
	my $opts_ok = Getopt::Long::GetOptions(
		'child'             => \$child,
		'forwardable|f'     => \$tkt_forwardable,
		'keytab|t=s'        => \$keytab,
		'lifetime|l=s'      => \$tkt_lifetime,
		'no-forwardable|F'  => \$tkt_not_forwardable,
		'parent-process=i'  => \$parent_pid,
		'pid-file=s'        => \$pid_file,
		'pid-timeout=i'     => \$pid_timeout,
		'principal|P=s'     => \$principal,
		'proxiable|p'       => \$tkt_proxiable,
		'renewable|r=s'     => \$tkt_renewable_lifetime,
		'verbose|v+'        => \$verbose,
	);
	if (!$opts_ok) {
		die "Error processing options";
	}

	if ($verbose >= 3) {
		print STDERR ('$FindBin::Bin is ', $FindBin::Bin, "\n");
		print STDERR ('$FindBin::RealBin is ', $FindBin::RealBin, "\n");
		print STDERR ('$FindBin::Script is ', $FindBin::Script, "\n");
		print STDERR ('$FindBin::RealScript is ', $FindBin::RealScript, "\n");
	}
	if ($verbose >= 2) {
		print STDERR ('After option processing, @ARGV is: ',
		              join(', ', @ARGV), "\n");
	}

	if (!defined($pid_file)) {
		die "The '--pid-file' option is required";
	}

	if ($child && !defined($parent_pid)) {
		die "The '--parent-process' option is required when '--child' is " .
		    "specified";
	}
	if (!$child && defined($parent_pid)) {
		die "The '--parent-process' option can only be specified in " .
		    "conjunction with '--child'";
	}

	if (!$child && $tkt_forwardable && $tkt_not_forwardable) {
		die "The '--forwardable/-f' option is mutually exclusive with " .
		    "the '--no-forwardable/-F' option";
	}
	if (!$child && !defined($principal)) {
		die "The '--principal/-P' option is required";
	}

	if (!$child) {
		parent_handling();
	} else {
		child_handling();
	}
	die "Internal error: handling function returned";
}

############################################################################
main();
